#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2019 Alan Carter. All rights reserved.
#

################################################################################
# Imports
################################################################################
import json
import requests
import urllib2
import time
import datetime
import re
import sys
import math
import threading

################################################################################
# Globals
################################################################################

########################################
def updateVar(name, value, folder=0):
    if name not in indigo.variables:
        indigo.variable.create(name, value=value, folder=folder)
    else:
        indigo.variable.updateValue(name, value)

################################################################################
class Plugin(indigo.PluginBase):
    ########################################
    # Class properties
    ########################################
    
    ########################################
    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
        self.userName = self.pluginPrefs.get("userName", "")
        self.userPass = self.pluginPrefs.get("userPass", "")
        self.sessionId = self.pluginPrefs.get("sessionId", "")
        self.loggingTherm = self.pluginPrefs.get("LogChoiceTherm", False)
        self.defaultBoostTemp = self.pluginPrefs.get("DefaultBoostTemp", 20)
        self.useAppDefault = self.pluginPrefs.get("UseAppDefault", False)
        self.defaultBoostHours = self.pluginPrefs.get("DefaultBoostHours", 1)
        self.limescaleProtect = self.pluginPrefs.get("LimescaleProtect", False)
        self.lpComplete = True
        self.lpInProgress = False
        self.lpClosePhase = False
        self.lpRequest = False
        self.lpCloseTime = 0
        self.showSetpointInNotes = self.pluginPrefs.get("ShowSetpointInNotes", False)
        self.debugEnabled = self.pluginPrefs.get("debugEnabled", False)
        self.timeout = 5
        self.homeId = ""
        self.trvList = []
        self.scheduleList = []
        self.max_schedules = 0
        self.commsOK = True
        self.thermOK = True
        self.lastUpdateTimeTherm = 0
        self.openList = {}
        self.hiveUrl = "https://api-prod.bgchprod.info:443/omnia/"
        self.contentType = "application/vnd.alertme.zoo-6.4+json"
        self.accept = "application/vnd.alertme.zoo-6.4+json"
        self.client = "Hive Web Dashboard"
        self.lastSessionTime = 0


    ########################################
    def __del__(self):
        indigo.PluginBase.__del__(self)
        
    ########################################
    def startup(self):
        indigo.server.log("Starting Hive plugin")
        self.firstTime = True

    def deviceStartComm(self, dev):
        dev.stateListOrDisplayStateIdChanged()
        return

    def shutdown(self):
        indigo.server.log("Stopping Hive plugin")
        

    ########################################
    def runConcurrentThread(self):
        if self.firstTime:
            self.userName = self.pluginPrefs.get("userName", "")
            self.userPass = self.pluginPrefs.get("userPass", "")
            self.sessionId = self.pluginPrefs.get("sessionId", "")
            self.loggingTherm = self.pluginPrefs.get("LogChoiceTherm", False)
            self.defaultBoostTemp = self.pluginPrefs.get("DefaultBoostTemp", 20)
            self.useAppDefault = self.pluginPrefs.get("UseAppDefault", False)
            self.defaultBoostHours = self.pluginPrefs.get("DefaultBoostHours", 1)
            self.defaultBoostPeriod = 60
            self.limescaleProtect = self.pluginPrefs.get("LimescaleProtect", False)
            self.lpComplete = True
            self.lpInProgress = False
            self.lpClosePhase = False
            self.lpRequest = False
            self.lpCloseTime = 0    
            self.showSetpointInNotes = self.pluginPrefs.get("ShowSetpointInNotes", False)
            self.debugEnabled = self.pluginPrefs.get("debugEnabled", False)
            self.createDevices()
            self.firstTime = False

        indigo.server.log("********************************************************************")
        indigo.server.log("Starting Hive monitoring thread")
        if self.limescaleProtect:
            indigo.server.log(" - Limescale protection will be scheduled weekly")
        else:
            indigo.server.log(" - Limescale protection schedule disabled")
        if self.loggingTherm:
            indigo.server.log(" - Thermostat message logging is on")
        indigo.server.log("********************************************************************")

        try:
            while True:
                if self.sessionId == "":
                    status, detail = self.auth()
                    if status == "error" and self.commsOK:
                        indigo.server.log("runConcurrentThread() - authentication failed.  Will keep trying.", isError=True)
                        self.commsOK = False
                        self.sleep(30)
                    elif status != "error" and self.commsOK == False:
                        indigo.server.log("runConcurrentThread() - authentication now OK")
                        self.commsOK = True
                if self.commsOK:
                    self.getUpdate()
                    self.sleep(60)
                
        except self.StopThread:
            self.shutdown()


    ######################################
    # Called when user prefs are changed   
            
    def closedPrefsConfigUi(self, valuesDict, userCancelled):
        # If the user saved preferences, update logging parameters
        if userCancelled == False:
            oldLoggingTherm = self.loggingTherm
            self.loggingTherm = valuesDict.get("LogChoiceTherm", False)
            if self.loggingTherm != oldLoggingTherm:
                if self.loggingTherm:
                    indigo.server.log("Logging all thermostat messages to Event Log")
                else:
                    indigo.server.log("Thermostat message logging is off")
            oldDefault = self.useAppDefault
            self.useAppDefault = valuesDict.get("UseAppDefault", False)
            if self.useAppDefault != oldDefault:
                if self.useAppDefault:
                    indigo.server.log("App default period will be used for overrides")
                else:
                    indigo.server.log("Plugin default period will be used for overrides")
            oldSpin = self.showSetpointInNotes
            self.showSetpointInNotes = valuesDict.get("ShowSetpointInNotes", False)
            if self.showSetpointInNotes != oldSpin:
                if self.showSetpointInNotes:
                    indigo.server.log("Setpoints will be shown in Indigo Home Window Notes field")
                else:
                    indigo.server.log("Setpoints will not be shown in Indigo Home Window Notes field")
            oldLime = self.limescaleProtect
            self.limescaleProtect = valuesDict.get("LimescaleProtect", False)
            if self.limescaleProtect != oldLime:
                if self.limescaleProtect:
                    indigo.server.log("Limescale protection will be scheduled weekly")
                else:
                    indigo.server.log("Limescale protection schedule disabled")
            oldDebug = self.debugEnabled
            self.debugEnabled = valuesDict.get("debugEnabled", False)
            if self.debugEnabled != oldDebug:
                if self.debugEnabled:
                    indigo.server.log("Debug messages will be written to Event Log")
                else:
                    indigo.server.log("Debug messages will not be written to Event Log")


        
    ########################################
    # Netatmo specific functions
    
    def createDevices(self):
        if self.userName == "":
            indigo.server.log("Missing User Name")   
        if self.userPass == "":
            indigo.server.log("Missing User Password")
        if self.sessionId == "":
            indigo.server.log("Renewing Session ID")
            status, detail = self.auth()
            if status == "error" and self.commsOK:
                indigo.server.log("Authentication failed in createDevices", isError=True)
                #indigo.server.log(detail, isError=True)
                self.commsOK = False
                #return
            elif status == "OK" and self.commsOK == False:
                indigo.server.log("Authentication in createDevices now OK")
                self.commsOK = True
        else:
            indigo.server.log("Current Session ID is valid; authentication not required")
            
        if self.commsOK:
            try:
                dataC = self.getNodes()
                zoneName = ""
                if dataC == "" or dataC == "error":
                    return
                max_nodes = len(dataC["nodes"])
                for node in range(0, max_nodes):
                    nodeType = dataC["nodes"][node]["nodeType"]
                    if nodeType == "http://alertme.com/schema/json/node.class.thermostat.json#":
                    	dataL = str(dataC["nodes"][node])
                        if 'zoneName' in dataL:
                            # Receiver
                            zoneName = dataC["nodes"][node]["name"]
                            continue
                        elif 'supportsHotWater' in dataL and zoneName != "":
                            if str(dataC["nodes"][node]["attributes"]["supportsHotWater"]["reportedValue"]) == "False":
                                # Heating
                                devName = zoneName+" Heating"
                                devId = dataC["nodes"][node]["id"]
                                device = None
                                for dev in indigo.devices.iter("self"):
                                    if dev.address == devId:
                                        device = indigo.devices[dev.name]
                                if device == None:
                                    indigo.server.log("Creating Hive device for %s" % (devName))
                                    device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=devId, name=devName, pluginId="com.racarter.indigoplugin.hive", deviceTypeId="hiveTherm", props={"SupportsBatteryLevel":True, "SupportsCoolSetpoint":False, "SupportsHvacFanMode":False})
                            elif str(dataC["nodes"][node]["attributes"]["supportsHotWater"]["reportedValue"]) == "True":
                                # Hot Water
                                devName = zoneName+" Hot Water"
                                devId = dataC["nodes"][node]["id"]
                                device = None
                                for dev in indigo.devices.iter("self"):
                                    if dev.address == devId:
                                        device = indigo.devices[dev.name]
                                if device == None:
                                    indigo.server.log("Creating Hive device for %s" % (devName))
                                    device = indigo.device.create(protocol=indigo.kProtocol.Plugin, address=devId, name=devName, pluginId="com.racarter.indigoplugin.hive", deviceTypeId="hiveHw", props={})
    #                       break


            except requests.exceptions.HTTPError as error:
                indigo.server.log("Thermostat device creation error: "+self.getErrorText(error.response.text), isError=True)
             
            except requests.Timeout as error:
                indigo.server.log("Thermostat device creation timeout error", isError=True)
                   

    def getUpdate(self):
        if (self.sessionId == "") or (time.time() - self.lastSessionTime > 1800):
            status, detail = self.auth()
            if status == 'error' and self.commsOK:
                indigo.server.log("Authentication failed in getUpdate():", isError=True)
                #indigo.server.log(detail, isError=True)
                self.commsOK = False
            elif status == "OK" and self.commsOK == False:
                indigo.server.log("Authentication in getUpdate() now OK")
                self.commsOK = True

        if self.commsOK:                    
            self.getThermUpdate()
#                 
#             if self.therm and (self.limescaleProtect or self.lpRequest):
#                 thisDay = datetime.datetime.today().weekday()   #Sunday = 6
#                 thisHour = datetime.datetime.now().hour
#                 if (self.lpRequest == True):
#                     self.limescaleProtectGo()
#                 elif (thisDay == 6) and (thisHour == 0) and (self.lpComplete == True):
#                     self.lpComplete = False                
#                 elif (thisDay == 6) and (thisHour == 15) and (self.lpComplete == False):
#                     self.limescaleProtectGo()
#                     
#             # Check for TRVs in limescale protection mode and open if necessary
#             if len(self.openList) > 0:
#                 for k, v in self.openList.items():
#                     if v < time.time():
#                         self.trvOpen(k)
#                         del self.openList[k]
#                         self.sleep(2)
                                  

    def getThermUpdate(self):
        dataU = self.getNodes()
        if dataU == "" or dataU == "error":
            return
        if self.loggingTherm:
            indigo.server.log("Hive data:")
            indigo.server.log(json.dumps(dataU, indent=4, sort_keys=True))
        try:
            max_nodes = len(dataU["nodes"])
        except:
            indigo.server.log("No data", isError=True)
            return
        for node in range(0, max_nodes):
            nodeType = dataU["nodes"][node]["nodeType"]
            if nodeType == "http://alertme.com/schema/json/node.class.thermostatui.json#":
                try:
                    batteryLevel = dataU["nodes"][node]["attributes"]["batteryLevel"]["reportedValue"]
                    batteryState = dataU["nodes"][node]["attributes"]["batteryState"]["reportedValue"]
                    break
                except:
                    indigo.server.log("No battery data", isError=True)
                    break
        for node in range(0, max_nodes):
            nodeType = dataU["nodes"][node]["nodeType"]
            if nodeType == "http://alertme.com/schema/json/node.class.thermostat.json#":
            	if "zoneName" in dataU["nodes"][node]:
            		continue
                try:
                	dataL = str(dataU["nodes"][node])
                	if 'supportsHotWater' in dataL:
						devId = dataU["nodes"][node]["id"]
						devName = dataU["nodes"][node]["name"]
						device = None
						for dev in indigo.devices.iter("self"):
							if dev.address == devId:
								device = indigo.devices[dev.name]
						if device != None:
							if str(dataU["nodes"][node]["attributes"]["supportsHotWater"]["reportedValue"]) == "False":
								# Heating
								temp = dataU["nodes"][node]["attributes"]["temperature"]["reportedValue"]
								sp = dataU["nodes"][node]["attributes"]["targetHeatTemperature"]["reportedValue"]
								mode = dataU["nodes"][node]["attributes"]["activeHeatCoolMode"]["reportedValue"]
								lock = dataU["nodes"][node]["attributes"]["activeScheduleLock"]["reportedValue"]
								heat = dataU["nodes"][node]["attributes"]["stateHeatingRelay"]["reportedValue"]
								if mode == "HEAT":
									modeKey = indigo.kHvacMode.ProgramHeat
									if str(lock) == "True":
										mode = "MANUAL"
									else:
										mode = "SCHEDULE"
								elif mode == "BOOST":
									modeKey = indigo.kHvacMode.Heat
								elif mode == "OFF":
									modeKey = indigo.kHvacMode.Off
								else:
									modeKey = indigo.kHvacMode.Off
									indigo.server.log("Mode error: "+mode, isError=True)
								now = datetime.datetime.now()
								dt_object = now.strftime("%Y-%m-%d %H:%M")
								keyValueList = [
									{'key':'temperatureInput1', 'value': temp, 'uiValue': str(temp)+"°C"},
									{'key':'setpointHeat', 'value':sp, 'uiValue': str(sp)+"°C"},
									{'key':'hvacOperationMode', 'value': modeKey},
									{'key':'batteryLevel', 'value':batteryLevel},
									{'key':'Mode', 'value':mode},
									{'key':'BatteryState', 'value':batteryState},
									{'key':'LastUpdate', 'value':dt_object},
								]
								device.updateStatesOnServer(keyValueList)
								if heat == "ON":
									device.updateStateImageOnServer(indigo.kStateImageSel.HvacHeating)
								else:
									device.updateStateImageOnServer(indigo.kStateImageSel.HvacHeatMode)
								if self.showSetpointInNotes:
									device.description = "{:>6}".format(str(sp)+"°C")
									device.replaceOnServer()
								elif device.description != "":
									device.description = ""
									device.replaceOnServer()
							elif str(dataU["nodes"][node]["attributes"]["supportsHotWater"]["reportedValue"]) == "True":
								# Hot Water
								heat = dataU["nodes"][node]["attributes"]["stateHotWaterRelay"]["reportedValue"]
								mode = dataU["nodes"][node]["attributes"]["activeHeatCoolMode"]["reportedValue"]
								lock = dataU["nodes"][node]["attributes"]["activeScheduleLock"]["reportedValue"]
								now = datetime.datetime.now()
								dt_object = now.strftime("%Y-%m-%d %H:%M")
								if heat == "ON":
									device.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
									onOff = True
								else:
									device.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
									onOff = False
								if mode == "HEAT":
									if str(lock) == "True":
										mode = "MANUAL"
									else:
										mode = "SCHEDULE"
								keyValueList = [
									{'key':'Mode', 'value':mode},
									{'key':'onOffState', 'value':onOff},
									{'key':'LastUpdate', 'value':dt_object},
								]
								device.updateStatesOnServer(keyValueList)

                except (requests.Timeout, requests.exceptions.HTTPError) as error:
                    if self.thermOK:
                        e = self.getErrorText(error.response.text)
                        indigo.server.log("Hive update timeout: %s" % (e), isError=True)
                        self.thermOK = False                 
                except:
                  if self.thermOK:
                      e = sys.exc_info()[0]
                      f = sys.exc_info()[1]
                      indigo.server.log("Hive update error: %s, %s" % (e, f), isError=True)
                      self.thermOK = False


    def limescaleProtectGo(self):
        # Drives TRVs closed/open for 5 minutes to protect against limescale build-up
        iterator = indigo.devices.iter(filter="self.netatmoTherm")
        for dev in iterator:
            if indigo.devices[dev.id].states["DeviceType"] == "TRV":
                indigo.server.log("%s limescale protection sequence initiated" % dev.name) 
                roomId = self.locations[dev.address]
                self.trvClose(roomId)
                now = int(time.time())
                openTime = now + 360
                self.openList[roomId] = openTime                
                self.sleep(5)
        self.lpRequest = False
        self.lpComplete = True


    def trvClose(self, roomId):
        # Used by limescale protect functions
        holdTemp = "8"
        now = int(time.time())
        endTime = str(now + 300)
        payload = {'access_token': self.accessToken, 'home_id': self.homeId, 'room_id': roomId, 'mode': 'manual', 'temp': holdTemp, 'endtime': endTime}
        try:
            response = requests.post("https://api.netatmo.com/api/setroomthermpoint", data=payload, timeout=self.timeout)
            response.raise_for_status()
            return True
        except:
            indigo.server.log("Limescale protection close TRV error", isError=True)
            return False


    def trvOpen(self, roomId):
        # Used by limescale protect functions
        holdTemp = "30"
        now = int(time.time())
        endTime = str(now + 300)
        payload = {'access_token': self.accessToken, 'home_id': self.homeId, 'room_id': roomId, 'mode': 'manual', 'temp': holdTemp, 'endtime': endTime}
        try:
            response = requests.post("https://api.netatmo.com/api/setroomthermpoint", data=payload, timeout=self.timeout)
            response.raise_for_status()
            return True
        except:
            indigo.server.log("Limescale protection open TRV error", isError=True)
            return False
        

    def listGeneratorA(self, filter="", valuesDict=None, typeId="", targetId=0):
        returnArray = []
        for sched in range(0, self.max_schedules):
            name = self.scheduleList[sched]
            returnArray.append(name)
        return sorted(returnArray,key=lambda x: x[1])
        
        
    def listGeneratorB(self, filter="", valuesDict=None, typeId="", targetId=0):
        returnArray = []
        iterator = indigo.devices.iter(filter="self.netatmoTherm")
        for dev in iterator:
            if indigo.devices[dev.id].states["DeviceType"] == "TRV":
                returnArray.append(dev.name)
        return sorted(returnArray,key=lambda x: x[1])
        

    def getErrorText(self, html):
        #Extract error text from html error message
        i = html.find("<title>")
        if i != -1:
            j = html.find("</title>")
            return html[i+7:j]
        else:
            return "Unknown error"
            
    
    def auth(self):
        try:
            url = self.hiveUrl+"auth/sessions"
            headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client}
            body = {
                "sessions": [
                {"username": self.userName, "password": self.userPass, "caller": "WEB"}]
            }
            response = requests.post(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
            response.raise_for_status()
            dataT = response.json()
            self.sessionId = dataT["sessions"][0]["sessionId"]
            self.lastSessionTime = time.time()
            return "OK", response.status_code 
            
        except requests.Timeout as error:
            indigo.server.log("Hive authorisation timeout", isError=True)
            indigo.server.log("Will try again in 1 minute")
            self.sleep(60)
            return "error"
                
        except:
          e = sys.exc_info()[0]
          f = sys.exc_info()[1]
          indigo.server.log("Hive authorisation error: %s, %s" % (e, f), isError=True)
          indigo.server.log("Will try again in 1 minute")
          self.sleep(60)
          return "error"


    def getNodes(self):
    	dataN = ""
        try:
			url = self.hiveUrl+"nodes"
			headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
			response = requests.get(url, headers=headers, timeout=self.timeout)
			response.raise_for_status()
			dataN = response.json()
			return dataN
            
        except requests.Timeout as error:
            indigo.server.log("Hive getNodes timeout", isError=True)
            return "error"
                
        except:
        	try:
        		status, detail = self.auth()
        	except:
          		e = sys.exc_info()[0]
          		f = sys.exc_info()[1]
          		indigo.server.log("Hive getNodes error: %s, %s" % (e, f), isError=True)
          		return "error"

    

    
    ########################################
    # Menu Item functions
    ######################


    def limescaleProtectNow(self):
        indigo.server.log("Initiating TRV Limescale Protection")
        self.lpRequest = True

    


    #######################################
    # Action functions
    ######################


    def setOverrideTherm(self, pluginAction):
        holdTemp = str(pluginAction.props["overrideTemp"])
        holdHours = pluginAction.props["numberOfHours"]
        holdMins = int(holdHours) * 60
        devToProcess = pluginAction.deviceId
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
        if device != None:
			try:
				url = self.hiveUrl+"nodes/"+device.address
				headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
				body = {"nodes": [{"attributes": {"activeHeatCoolMode": {"targetValue": "BOOST"},"scheduleLockDuration": {"targetValue": holdMins},"targetHeatTemperature": {"targetValue": holdTemp}}}]}
				response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
				response.raise_for_status()
				indigo.server.log("%s boost set at %s degC for %s minutes" % (device.name, str(holdTemp), str(holdMins)))
				if self.showSetpointInNotes:
					dev.description = "{:>6}".format(str(holdTemp)+"°C")
					dev.replaceOnServer()
				elif dev.description != "":
					dev.description = ""
					dev.replaceOnServer()         
			except requests.Timeout as error:
				e = self.getErrorText(error.response.text)
				indigo.server.log("Hive set boost timeout: %s" % (e), isError=True)              
			except:
			  e = sys.exc_info()[0]
			  f = sys.exc_info()[1]
			  indigo.server.log("Hive set boost error: %s, %s" % (e, f), isError=True)
			  
    def setOverrideHw(self, pluginAction):
        holdHours = pluginAction.props["numberOfHours"]
        holdMins = int(holdHours) * 60
        devToProcess = pluginAction.deviceId
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
        if device != None:
			try:
				url = self.hiveUrl+"nodes/"+device.address
				headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
				body = {"nodes": [{"attributes": {"activeHeatCoolMode": {"targetValue": "BOOST"},"scheduleLockDuration": {"targetValue": holdMins}}}]}
				response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
				response.raise_for_status()
				indigo.server.log("%s boost set for %s minutes" % (device.name, str(holdMins)))    
			except requests.Timeout as error:
				e = self.getErrorText(error.response.text)
				indigo.server.log("Hive set HW boost timeout: %s" % (e), isError=True)              
			except:
			  e = sys.exc_info()[0]
			  f = sys.exc_info()[1]
			  indigo.server.log("Hive set HW boost error: %s, %s" % (e, f), isError=True)
                

    def cancelOverride(self, pluginAction):
        devToProcess = pluginAction.deviceId
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.id == devToProcess:
                device = indigo.devices[dev.name]
        if device != None:
			try:
				url = self.hiveUrl+"nodes/"+device.address
				headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
				body = {"nodes": [{"attributes": {"activeHeatCoolMode": {"targetValue": "BOOST"},"scheduleLockDuration": {"targetValue": 0}}}]}
				response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
				response.raise_for_status()
				indigo.server.log("%s boost cancelled" % device.name)       
			except requests.Timeout as error:
				e = self.getErrorText(error.response.text)
				indigo.server.log("Hive cancel boost timeout: %s" % (e), isError=True)              
			except:
			  e = sys.exc_info()[0]
			  f = sys.exc_info()[1]
			  indigo.server.log("Hive cancel boost error: %s, %s" % (e, f), isError=True)


    def lsProtect(self, pluginAction):
        # Drives TRVs open/closed for 5 minutes to protect against limescale build-up
        indigo.server.log("Function not yet implemented", isError=True)
        return
        
        devToProcess = pluginAction.props.get("devId", "")
        indigo.server.log("%s limescale protection sequence initiated" % devToProcess)
        device = None
        for dev in indigo.devices.iter("self"):
            if dev.name == devToProcess:
                device = indigo.devices[dev.name]
        if device != None:
            roomId = self.locations[device.address]
            self.trvClose(roomId)
            # Add TRV to list of valves to be opened after time delay
            openTime = now + 360
            self.openList[roomId] = openTime  
            return True
        else:
            indigo.server.log("Limescale protection: device not found", isError=True)


                
    ########################################
    # Netatmo Thermostat/TRV Action callback
    ######################
    # Main thermostat action bottleneck called by Indigo Server.

    def actionControlThermostat(self, action, dev):
        if action.thermostatAction == indigo.kThermostatAction.SetHeatSetpoint:
            newSetpoint =  round(action.actionValue * 2) / 2
            if newSetpoint != action.actionValue:
                warning = " (Setpoint rounded to nearest half degree)"
            else:
                warning = ""
            try:
                url = self.hiveUrl+"nodes/"+dev.address
                headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
                body = {"nodes": [{"attributes": {"targetHeatTemperature": {"targetValue": newSetpoint}}}]}
                response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
                response.raise_for_status()
                indigo.server.log(dev.name+" setpoint changed to "+str(newSetpoint)+" degC"+warning)
                dev.updateStateOnServer(key="setpointHeat", value=newSetpoint, uiValue=str(newSetpoint)+"°C")
                if self.showSetpointInNotes:
                    dev.description = "{:>6}".format(str(newSetpoint)+"°C")
                    dev.replaceOnServer()
                elif dev.description != "":
                    dev.description = ""
                    dev.replaceOnServer()         
            except requests.Timeout as error:
                e = self.getErrorText(error.response.text)
                indigo.server.log("Hive setpoint change timeout: %s" % (e), isError=True)              
            except:
              e = sys.exc_info()[0]
              f = sys.exc_info()[1]
              indigo.server.log("Hive setpoint change error: %s, %s" % (e, f), isError=True)
            
        elif action.thermostatAction == indigo.kThermostatAction.DecreaseHeatSetpoint:
            newSetpoint = dev.heatSetpoint - action.actionValue
            if newSetpoint != dev.heatSetpoint:
				try:
					url = self.hiveUrl+"nodes/"+dev.address
					headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
					body = {"nodes": [{"attributes": {"targetHeatTemperature": {"targetValue": newSetpoint}}}]}
					response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
					response.raise_for_status()
					indigo.server.log(dev.name+" setpoint decreased to "+str(newSetpoint)+" degC")
					dev.updateStateOnServer(key="setpointHeat", value=newSetpoint, uiValue=str(newSetpoint)+"°C")
					if self.showSetpointInNotes:
						dev.description = "{:>6}".format(str(newSetpoint)+"°C")
						dev.replaceOnServer()
					elif dev.description != "":
						dev.description = ""
						dev.replaceOnServer()         
				except requests.Timeout as error:
					e = self.getErrorText(error.response.text)
					indigo.server.log("Hive setpoint decrease timeout: %s" % (e), isError=True)              
				except:
				  e = sys.exc_info()[0]
				  f = sys.exc_info()[1]
				  indigo.server.log("Hive setpoint decrease error: %s, %s" % (e, f), isError=True)
        
        elif action.thermostatAction == indigo.kThermostatAction.IncreaseHeatSetpoint:
            newSetpoint = dev.heatSetpoint + action.actionValue
            if newSetpoint != dev.heatSetpoint:
                try:
                    url = self.hiveUrl+"nodes/"+dev.address
                    headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
                    body = {"nodes": [{"attributes": {"targetHeatTemperature": {"targetValue": newSetpoint}}}]}
                    response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
                    response.raise_for_status()
                    indigo.server.log(dev.name+" setpoint increased to "+str(newSetpoint)+" degC")
                    dev.updateStateOnServer(key="setpointHeat", value=newSetpoint, uiValue=str(newSetpoint)+"°C")
                    if self.showSetpointInNotes:
                        dev.description = "{:>6}".format(str(newSetpoint)+"°C")
                        dev.replaceOnServer()
                    elif dev.description != "":
                        dev.description = ""
                        dev.replaceOnServer()         
                except requests.Timeout as error:
                    e = self.getErrorText(error.response.text)
                    indigo.server.log("Hive setpoint increase timeout: %s" % (e), isError=True)              
                except:
                  e = sys.exc_info()[0]
                  f = sys.exc_info()[1]
                  indigo.server.log("Hive setpoint increase error: %s, %s" % (e, f), isError=True)
                  
        elif action.thermostatAction == indigo.kThermostatAction.SetHvacMode:
            if action.actionMode == indigo.kHvacMode.Off:
                try:
                    url = self.hiveUrl+"nodes/"+dev.address
                    headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
                    body = {"nodes": [{"attributes": {"activeHeatCoolMode": {"targetValue": "OFF"}}}]}
                    response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
                    response.raise_for_status()
                    dev.updateStateOnServer(key="hvacOperationMode", value=indigo.kHvacMode.Off)
                    indigo.server.log("Thermostat switched OFF")       
                except requests.Timeout as error:
                    e = self.getErrorText(error.response.text)
                    indigo.server.log("Hive switch off timeout: %s" % (e), isError=True)              
                except:
                  e = sys.exc_info()[0]
                  f = sys.exc_info()[1]
                  indigo.server.log("Hive switch off error: %s, %s" % (e, f), isError=True)
            elif action.actionMode == indigo.kHvacMode.Heat:
                try:
                    url = self.hiveUrl+"nodes/"+dev.address
                    headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
                    body = {'nodes': [{ 'attributes':{"activeHeatCoolMode": {"targetValue": "HEAT",},"activeScheduleLock": {"targetValue": "true"}}}]}
                    response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
                    response.raise_for_status()
                    dev.updateStateOnServer(key="hvacOperationMode", value=indigo.kHvacMode.Heat)
                    indigo.server.log("Thermostat switched to MANUAL")      
                except requests.Timeout as error:
                    e = self.getErrorText(error.response.text)
                    indigo.server.log("Hive switch to manual timeout: %s" % (e), isError=True)              
                except:
                  e = sys.exc_info()[0]
                  f = sys.exc_info()[1]
                  indigo.server.log("Hive switch to manual error: %s, %s" % (e, f), isError=True)
            elif action.actionMode == indigo.kHvacMode.HeatCool:
                try:
                    url = self.hiveUrl+"nodes/"+dev.address
                    headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
                    body = {'nodes': [{ 'attributes':{"activeHeatCoolMode": {"targetValue": "HEAT",},"activeScheduleLock": {"targetValue": "false"}}}]}
                    response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
                    response.raise_for_status()
                    dev.updateStateOnServer(key="hvacOperationMode", value=indigo.kHvacMode.ProgramHeat)
                    indigo.server.log("Thermostat switched to SCHEDULE") 
                except requests.Timeout as error:
                    e = self.getErrorText(error.response.text)
                    indigo.server.log("Hive switch to schedule timeout: %s" % (e), isError=True)              
                except:
                  e = sys.exc_info()[0]
                  f = sys.exc_info()[1]
                  indigo.server.log("Hive switch to schedule error: %s, %s" % (e, f), isError=True)
            else:
                indigo.server.log("Invalid mode command.  Supported modes are 'Off', 'Heat On' (= MANUAL), 'Auto Heat/Cool' (= SCHEDULE)", isError=True)


    def actionControlDevice(self, action, dev):
    	if action.deviceAction == indigo.kDeviceAction.TurnOn:
			try:
				url = self.hiveUrl+"nodes/"+dev.address
				headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
				body = {"nodes": [{"attributes": {"activeHeatCoolMode": {"targetValue": "HEAT"}}}]}
				response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
				response.raise_for_status()
				dev.updateStateOnServer("onOffState", True)      
			except requests.Timeout as error:
				e = self.getErrorText(error.response.text)
				indigo.server.log("Hive HW switch on timeout: %s" % (e), isError=True)              
			except:
			  e = sys.exc_info()[0]
			  f = sys.exc_info()[1]
			  indigo.server.log("Hive HW switch on error: %s, %s" % (e, f), isError=True)
    	elif action.deviceAction == indigo.kDeviceAction.TurnOff:
			try:
				url = self.hiveUrl+"nodes/"+dev.address
				headers = {'Content-Type': self.contentType, 'Accept': self.accept, 'X-Omnia-Client': self.client, 'X-Omnia-Access-Token': self.sessionId}
				body = {"nodes": [{"attributes": {"activeHeatCoolMode": {"targetValue": "OFF"}}}]}
				response = requests.put(url, headers=headers, data=json.dumps(body), timeout=self.timeout)
				response.raise_for_status()
				dev.updateStateOnServer("onOffState", False)      
			except requests.Timeout as error:
				e = self.getErrorText(error.response.text)
				indigo.server.log("Hive HW switch off timeout: %s" % (e), isError=True)              
			except:
			  e = sys.exc_info()[0]
			  f = sys.exc_info()[1]
			  indigo.server.log("Hive HW switch off error: %s, %s" % (e, f), isError=True)
    	else:
    		indigo.server.log("Unsupported device action: %s" % action.deviceAction, isError=True)

